/* dopewars.c    dopewars - general purpose routines and initialisation */ 
/* Copyright (C)  1998-9  Ben Webb                                      */
/*                Email: ben@bellatrix.pcl.ox.ac.uk                     */
/*                WWW: http://bellatrix.pcl.ox.ac.uk/~ben/dopewars/     */

/* This program is free software; you can redistribute it and/or        */
/* modify it under the terms of the GNU General Public License          */
/* as published by the Free Software Foundation; either version 2       */
/* of the License, or (at your option) any later version.               */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program; if not, write to the Free Software          */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               */
/*                   MA  02111-1307, USA.                               */

#include <stdio.h>
#include <ncurses.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include "dopewars.h"
#include "message.h"
#include "serverside.h"
#include "clientside.h"
#include "AIPlayer.h"

int ClientSock,ListenSock;     
char Network,Client,Server,AIPlayer;
/* dopewars acting as standalone TCP server:
             Network=Server=TRUE   Client=FALSE
   dopewars acting as client, connecting to standalone server:
             Network=Client=TRUE   Server=FALSE
   dopewars in single-player or antique mode:
             Network=Server=FALSE Client=TRUE
*/
int Port;
FILE *out;
char HiScoreFile[80],ServerName[80];
char WantHelp,WantVersion,WantAntique,WantColour,WantNetwork;
int NUMLOCATION;
struct PLAYER *Player,Noone;

char *Discover[NUMDISCOVER] = {
   "escaped", "defected", "was shot" };
char *Playing[NUMPLAYING] = {
   "`Are you Experienced` by Jimi Hendrix",
   "`Cheeba Cheeba` by Tone Loc",
   "`Comin` in to Los Angeles` by Arlo Guthrie",
   "`Commercial` by Spanky and Our Gang",
   "`Late in the Evening` by Paul Simon",
   "`Light Up` by Styx",
   "`Mexico` by Jefferson Airplane",
   "`One toke over the line` by Brewer & Shipley",
   "`The Smokeout` by Shel Silverstein",
   "`White Rabbit` by Jefferson Airplane",
   "`Itchycoo Park` by Small Faces",
   "`White Punks on Dope` by the Tubes",
   "`Legend of a Mind` by the Moody Blues",
   "`Eight Miles High` by the Byrds",
   "`Acapulco Gold` by Riders of the Purple Sage",
   "`Kicks` by Paul Revere & the Raiders",
   "the Nixon tapes",
   "`Legalize It` by Mojo Nixon & Skid Roper"
};


char *StoppedTo[NUMSTOPPEDTO] = {
   "have a beer",
   "smoke a joint",
   "smoke a cigar",
   "smoke a Djarum",
   "smoke a cigarette"
};

struct GUN Gun[NUMGUN] = {
   { "Baretta",3000,4,5 },
   { ".38 Special",3500,4,9 },
   { "Ruger",2900,4,4 },
   { "Saturday Night Special",3100,4,7 }
};

struct HISCORE MultiScore[NUMHISCORE],AntiqueScore[NUMHISCORE];

struct LOCATION Location[MAXLOCATION] = {
   { "Bronx",10,NUMDRUG/2+1,NUMDRUG },
   { "Ghetto",5,NUMDRUG/2+2,NUMDRUG },
   { "Central Park",15,NUMDRUG/2,NUMDRUG },
   { "Manhattan",90,NUMDRUG/2-2,NUMDRUG-2 },
   { "Coney Island",20,NUMDRUG/2,NUMDRUG },
   { "Brooklyn",70,NUMDRUG/2-2,NUMDRUG-1 },
   { "Queens",50,NUMDRUG/2,NUMDRUG },
   { "Staten Island",20,NUMDRUG/2,NUMDRUG }
};

struct DRUG Drug[NUMDRUG] = {
   { "Acid",1000,4400,1,0,0,0,0 },
   { "Cocaine",15000,29000,0,1,0,0,0 },
   { "Hashish",480,1280,1,0,0,0,0 },
   { "Heroin",5500,13000,0,1,0,0,0 },
   { "Ludes",11,60,1,0,0,0,0 },
   { "MDA",1500,4400,0,0,0,0,0 },
   { "Opium",540,1250,0,1,0,0,0 },
   { "PCP",1000,2500,0,0,0,0,0 },
   { "Peyote",220,700,0,0,0,0,0 },
   { "Shrooms",630,1300,0,0,0,0,0 },
   { "Speed",90,250,0,1,0,0,0 },
   { "Weed",315,890,1,0,0,0,0 }
};

char *SubwaySaying[NUMSUBWAY] = {
   "Wouldn\'t it be funny if everyone suddenly quacked at once?",
   "The Pope was once Jewish, you know",
   "I\'ll bet you have some really interesting dreams",
   "So I think I\'m going to Amsterdam this year",
   "Son, you need a yellow haircut",
   "I think it\'s wonderful what they\'re doing with incense these days",
   "I wasn\'t always a woman, you know",
   "Does your mother know you\'re a dope dealer?",
   "Are you high on something?",
   "Oh, you must be from California",
   "I used to be a hippie, myself",
   "There\'s nothing like having lots of money",
   "You look like an aardvark!",
   "I don\'t believe in Ronald Reagan",
   "Courage!  Bush is a noodle!",
   "Haven\'t I seen you on TV?",
   "I think hemorrhoid commercials are really neat!",
   "We\'re winning the war for drugs!",
   "A day without dope is like night",
   "We only use 20% of our brains, so why not burn out the other 80%",
   "I\'m soliciting contributions for Zombies for Christ",
   "I\'d like to sell you an edible poodle",
   "Winners don\'t do drugs... unless they do",
   "Kill a cop for Christ!",
   "I am the walrus!",
   "Jesus loves you more than you will know",
   "I feel an unaccountable urge to dye my hair blue",
   "Wasn\'t Jane Fonda wonderful in Barbarella",
   "Just say No... well, maybe... ok, what the hell!",
   "Would you like a jelly baby?",
   "Drugs can be your friend!"
};

int brandom(int bot,int top) {
/* Returns a random integer not less than bot and less than top */
   return (int)((float)(top-bot)*rand()/(RAND_MAX+1.0))+bot;
}

int GetKey(char *allowed) {
/* Waits for keyboard input; will only accept a key listed in the */
/* "allowed" string. Returns the key pressed. Case insensitive.   */
   int i,c;
   curs_set(1);
   c=0;
   if (!allowed || strlen(allowed)==0) return 0;
   while (1) {
      c=getch(); c=toupper(c);
      for (i=0;i<strlen(allowed);i++) if (allowed[i]==c) {
         addch(c | TextAttr);
         curs_set(0); return c;
      }
   }
   curs_set(0);
   return 0;
}

struct PLAYER *AddPlayer(int fd,struct PLAYER **First) {
/* Creates a new PLAYER structure, initialises it, and adds it to the */
/* linked list starting at "First". If this function is called by the */
/* server, then it should pass the file descriptor of the socket used */
/* to communicate with the client player. Returns a pointer to the    */
/* new structure.  */
   struct PLAYER *NewPlayer,*tmp;
   NewPlayer=(struct PLAYER *)malloc(sizeof(struct PLAYER));
   NewPlayer->fd=-1;
   NewPlayer->Next=NULL;
   NewPlayer->Name[0]=0;
   NewPlayer->IsAt=0;
   NewPlayer->Attacked=NULL;
   NewPlayer->EventNum=E_NONE;
   NewPlayer->Here = (DrugHere *)calloc(1,sizeof(DrugHere));
   InitList(&(NewPlayer->SpyList));
   InitList(&(NewPlayer->TipList));
   if (Server) {
      NewPlayer->fd=fd;
      NewPlayer->out=fdopen(fd,"w");
   } 
   if (*First==NULL) {
      *First=NewPlayer;
   } else {
      tmp=*First;
      while (tmp->Next) tmp=tmp->Next;
      tmp->Next=NewPlayer;
   }
   return NewPlayer;
}

void RemovePlayer(struct PLAYER *Play,struct PLAYER **First) {
/* Removes the PLAYER structure pointed to by "Play" from the linked */
/* list starting at "First". The client socket is freed if called    */
/* from the server.                                                  */
   struct PLAYER *tmp;
   if (*First==Play) {
      *First=Play->Next;
   } else {
      tmp=*First;
      while (tmp && tmp->Next != Play) tmp=tmp->Next;
      if (!tmp) return;
      tmp->Next=Play->Next;
   }  
   if (Server) {
      fclose(Play->out); close(Play->fd);
   }
   free(Play->Here);
   ClearList(&(Play->SpyList));
   ClearList(&(Play->TipList));
   free(Play);
}

void InitialisePlayer(struct PLAYER *Play) {
/* Initialises the fields of the PLAYER structure pointed to by "Play" */
   Play->Turn=1; Play->Cash=2000;
   Play->Debt=5500; Play->Bank=0;
   Play->Health=100;
   ClearInventory(&(Play->Carry));
   ClearInventory(&(Play->OnGround));
   Play->CoatSize=100; Play->IsAt=0;
   Play->Name[0]=0;
   Play->Bitches=8;
   Play->Flags=0;
}

struct PLAYER *GetPlayerByName(char *Name,struct PLAYER *First) {
/* Searches the linked list starting at "First" for a PLAYER structure */
/* with the name "Name". Returns a pointer to this structure, or NULL  */
/* if no match can be found.                                           */
   struct PLAYER *tmp;
   if (Name==NULL) return &Noone;
   if (Name[0]==0) return &Noone;
   for (tmp=First;tmp;tmp=tmp->Next) {
      if (strcmp(tmp->Name,Name)==0) return tmp;
   }
   return NULL;
}

char IsValidPlayer(struct PLAYER *Play,struct PLAYER *First) {
/* Returns TRUE if "Play" is within the linked list starting at "First" */
   struct PLAYER *tmp;
   if (!Play || !First) return FALSE;
   for (tmp=First;tmp;tmp=tmp->Next) {
      if (tmp==Play) return TRUE;
   }
   return FALSE;
}

void clear_line(int line) {
/* Clears one whole 80-character line on the curses screen */
   int i;
   move(line,0);
   for (i=0;i<80;i++) addch(' ');
}

void clear_exceptfor(int skip) {
/* Clears the bottom of the screen (i.e. from line 16 to line 23) */
/* except for the top "skip" lines                                */
   int i;
   for (i=16+skip;i<=23;i++) clear_line(i);
}


void clear_bottom() {
/* Clears screen lines 16 to 23 */
   int i;
   for (i=16;i<=23;i++) clear_line(i);
}

void clear_screen() {
/* Clears the entire screen; 24 lines of 80 characters each */
   int i;
   for (i=0;i<24;i++) clear_line(i);
}

void nice_wait() {
/* Displays a prompt on the bottom screen line and waits for the user */
/* to press a key                                                     */
   attrset(PromptAttr);
   mvaddstr(23,32,"Press any key...");
   getch();
   attrset(TextAttr);
}

void display_message(char *buf) {
/* Displays a message "buf" in the message area (lines 10 to 14) */
/* scrolling previous messages up                                */
   int x,y;
   chtype ch;
   attrset(StatsAttr);
   for (y=11;y<=14;y++) for (x=2;x<78;x++) {
      ch=mvinch(y,x);
      mvaddch(y-1,x,ch);
   }
   buf[76]=0;
   if (strlen(buf)<76) for (x=strlen(buf);x<76;x++) {
      buf[x]=' ';
   }
   buf[76]=0;
/*   scroll(MsgWin);*/
   mvaddstr(14,2,buf);
   refresh();
}

void start_curses() {
/* Initialises the curses library for accessing the screen */
   initscr();
   if (WantColour) {
      start_color();
      init_pair(1,COLOR_MAGENTA,COLOR_WHITE);
      init_pair(2,COLOR_BLACK,COLOR_WHITE);
      init_pair(3,COLOR_BLACK,COLOR_WHITE);
      init_pair(4,COLOR_BLUE,COLOR_WHITE);
      init_pair(5,COLOR_WHITE,COLOR_BLUE);
      init_pair(6,COLOR_RED,COLOR_WHITE);
   }
   cbreak();
   noecho();
   nodelay(stdscr,FALSE);
   keypad(stdscr,TRUE);
   curs_set(0);
}

void end_curses() {
/* Shuts down the curses screen library */
   keypad(stdscr,FALSE);
   curs_set(1);
   erase();
   refresh();
   endwin();
}

void nice_input(char *prompt,char *str,int maxlen,int sy,int sx,
                char digitsonly,char displaystr) {
/* Displays the given "prompt" (if non-NULL) at coordinates sx,sy and   */
/* allows the user to input a string, which is stored in "str". No more */
/* than "maxlen" characters will be read (not including terminating     */
/* NULL). If "digitsonly" is TRUE, the user will be permitted only to   */
/* input numbers; if "displaystr" is TRUE, the starting value of "str"  */
/* is displayed as the default response.                                */
   int i,c,x;
   x=sx;
   move(sy,x);
   if (prompt) {
      attrset(PromptAttr);
      addstr(prompt);
      x+=strlen(prompt);
   }
   attrset(TextAttr);
   if (displaystr) {
      addstr(str);
      i=strlen(str);
   } else i=0;
   curs_set(1);
   while(1) {
      move(sy,x+i);
      c=getch();
      if (c==KEY_ENTER || c=='\n') {
         str[i]=0;
         break;
      } else if ((c==8 || c==KEY_BACKSPACE || c==127) && i>0) {
         move(sy,x+i-1);
         addch(' ');
         i--;
      } else if (((digitsonly && c>='0' && c<='9') || 
                 (!digitsonly && c>=32 && c!='^' && c<127)) && i<maxlen) {
         str[i++]=c;
         addch(c);
      }
   }
   curs_set(0);
   move(sy,x);
}

void display_intro() {
/* Displays a dopewars introduction screen */
   char text[80];
   attrset(TextAttr);
   clear_screen();
   attrset(TitleAttr);
   mvaddstr(1,31,"D O P E   W A R S");

   attrset(TextAttr);

   mvaddstr(3,1,"Based on John E. Dell's old Drug Wars game, Dope Wars \
is a simulation of an");
   mvaddstr(4,1,"imaginary drug market.  Dope Wars is an All-American \
game which features");
   mvaddstr(5,1,"buying, selling, and trying to get past the cops!");

   mvaddstr(7,1,"The first thing you need to do is pay off your \
debt to the Loan Shark. After");
   mvaddstr(8,1,"that, your goal is to make as much money as \
possible (and stay alive)! You");
   mvaddstr(9,1,"have one month of game time to make your fortune.");

   mvaddstr(11,20,"Copyright (C) 1998/9  Ben Webb \
ben@bellatrix.pcl.ox.ac.uk");
   sprintf(text,"Version %d.%d.%s",VERSION,MAJVER,MINVER);
   mvaddstr(11,3,text);
   mvaddstr(12,12,"dopewars is released under the GNU General Public Licence");

   mvaddstr(14,7,"Drug Dealing and Research     Dan Wolf");
   mvaddstr(15,7,"Play Testing                  Phil Davis           \
Owen Walsh");
   mvaddstr(16,7,"Extensive Play Testing        Katherine Holt       \
Caroline Moore");
   mvaddstr(17,7,"Constructive Criticism        Andrea Elliot-Smith  \
Pete Winn");
   mvaddstr(18,7,"Unconstructive Criticism      James Matthews");

   mvaddstr(20,3,"For information on the command line options, type \
dopewars -h at your");
   mvaddstr(21,1,"Unix prompt. This will display a help screen, listing \
the available options.");

   nice_wait();
   attrset(TextAttr); clear_screen(); refresh();
}

void jet(struct PLAYER *Play) {
/* Displays the list of locations and prompts the user to select one. */
/* If the current location is selected, simply drop back to the main  */
/* game loop, otherwise send a request to the server to move to the   */
/* new location. The active client player is stored in "Play"         */
   int i,c;
   char text[80];
   clear_bottom();
   for (i=0;i<NUMLOCATION;i++) {
      sprintf(text,"%d. %s",i+1,Location[i].Name);
      mvaddstr(17+i/3,(i%3)*20+12,text);
   }
   attrset(PromptAttr);
   mvaddstr(22,22,"Where to, dude ? ");
   attrset(TextAttr);
   curs_set(1);
   while (1) {
      c=getch();
      if (c>='1' && c<'1'+NUMLOCATION) {
         addstr(Location[c-'1'].Name);
         if (Play->IsAt != c-'1') {
            for (i=0;i<NUMDRUG;i++) Play->Here->Price[i]=0;
            sprintf(text,"%d",c-'1');
            SendClientMessage(Play,C_NONE,"requestjet",NULL,text);
         }
         break;
      }
   }
   curs_set(0);
}

void DealDrugs(struct PLAYER *Play,char Buy) {
/* Prompts the user (i.e. the owner of client "Play") to buy drugs if   */
/* "Buy" is TRUE, or to sell drugs otherwise. A list of available drugs */
/* is displayed, and on receiving the selection, the user is prompted   */
/* for the number of drugs desired. Finally a message is sent to the    */
/* server to buy or sell the required quantity.                         */
   int c,NumDrugsHere;
   char text[100],input[80];
   int DrugNum,CanCarry,CanAfford;
   char DrugHere[NUMDRUG];

   NumDrugsHere=0;
   for (c=0;c<NUMDRUG;c++) if (Play->Here->Price[c]>0) {
      DrugHere[NumDrugsHere]=c;
      NumDrugsHere++;
   }

   clear_line(22);
   attrset(PromptAttr);
   mvaddstr(22,20,"What do you wish to ");
   addstr(Buy ? "buy? " : "sell? ");
   curs_set(1);
   attrset(TextAttr);
   while (1) {
      c=getch();
      c=toupper(c);
      if (c>='A' && c<'A'+NumDrugsHere) {
         DrugNum=DrugHere[c-'A'];
         addstr(Drug[DrugNum].Name);
         CanCarry=Play->CoatSize;
         CanAfford=Play->Cash/Play->Here->Price[DrugNum];

         if (Buy) {
            sprintf(text,"You can afford %d, and can carry %d. ",
                    CanAfford,CanCarry);
            mvaddstr(23,2,text);
            nice_input("How many do you buy? ",input,5,23,
                       2+strlen(text),1,0);
            c=atoi(input);
            if (c>=0) {
               sprintf(text,"drug^%d^%d",DrugNum,c);
               SendClientMessage(Play,C_NONE,"buyobject",NULL,text);
            }
         } else {
            sprintf(text,"You have %d. ",
                    Play->Carry.Drug[DrugNum]);
            mvaddstr(23,2,text);
            nice_input("How many do you sell? ",input,
                       5,23,2+strlen(text),1,0);
            c=atoi(input);
            if (c>=0) {
               sprintf(text,"drug^%d^%d",DrugNum,-c);
               SendClientMessage(Play,C_NONE,"buyobject",NULL,text);
            }
         }
         break;
      }
   }
   curs_set(0);
}

char *FormatPrice(long price,char *buffer) {
/* Takes the number in "price" and prints it into the string "buffer" */
/* adding commas to split up thousands, and adding a currency symbol  */
/* to the start. Returns a pointer to the string.                     */
   char text[80];
   text[0]=0;
   while (text[0]==0 || labs(price)>0) {
      if (labs(price)>=1000) sprintf(buffer,"%03ld",labs(price)%1000l);
      else sprintf(buffer,"%ld",price%1000l);
      price/=1000l;
      if (text[0]) strcat(buffer,",");      
      strcat(buffer,text);
      strcpy(text,buffer);
   }
   strcpy(buffer,"$");
   strcat(buffer,text);
   return buffer;
}

void print_location(char *text) {
/* Displays the string "text" at the top of the screen. Usually used for */
/* displaying the current location or the "Subway" flash.                */
   int i;
   attrset(LocationAttr);
   move(0,31);
   for (i=31;i<=49;i++) addch(' ');
   mvaddstr(0,40-strlen(text)/2,text);
   attrset(TextAttr);
}

int TotalGunsCarried(struct PLAYER *Player) {
/* Returns the total number of guns being carried by "Player" */
   int i,c;
   c=0;
   for (i=0;i<NUMGUN;i++) c+=Player->Carry.Gun[i];
   return c;
}

void print_status(struct PLAYER *Play,char DispDrug) {
/* Displays the status of player "Play" - i.e. the current turn, the   */
/* location, bitches, available space, cash, guns, health and bank     */
/* details. If "DispDrugs" is TRUE, displays the carried drugs on the  */
/* right hand side of the screen; if FALSE, displays the carried guns. */
   char text[80],num[80];
   int i,c;
   attrset(TitleAttr);
   sprintf(text,"12-%02d-1984",Play->Turn);
   mvaddstr(0,3,text);

   attrset(StatsAttr);
   for (i=2;i<=14;i++) {
      mvaddch(i,1,ACS_VLINE);
      mvaddch(i,78,ACS_VLINE);
   }
   mvaddch(1,1,ACS_ULCORNER);
   for (i=0;i<76;i++) addch(ACS_HLINE);
   addch(ACS_URCORNER);

   mvaddch(1,40,ACS_TTEE);
   for (i=2;i<=(WantAntique ? 13 : 8);i++) {
      move(i,2);
      for (c=2;c<40;c++) addch(' ');
      addch(ACS_VLINE);
      for (c=41;c<78;c++) addch(' ');
   }
   if (WantAntique) {
      mvaddch(14,1,ACS_LLCORNER);
      for (i=0;i<76;i++) addch(ACS_HLINE);
      addch(ACS_LRCORNER);
      mvaddch(14,40,ACS_BTEE);
   } else {
      mvaddch(9,1,ACS_LTEE);
      for (i=0;i<76;i++) addch(ACS_HLINE);
      addch(ACS_RTEE);
      mvaddstr(9,15,"Messages");
      mvaddch(9,40,ACS_BTEE);
      mvaddch(15,1,ACS_LLCORNER);
      for (i=0;i<76;i++) addch(ACS_HLINE);
      addch(ACS_LRCORNER);
   }

   mvaddstr(1,15,"Stats");

   attrset(StatsAttr);
   sprintf(text,"Cash %15s",FormatPrice(Play->Cash,num));
   mvaddstr(3,10,text);
   sprintf(text,"Guns             %3d",TotalGunsCarried(Play));
   mvaddstr(WantAntique ? 5 : 4,10,text);
   sprintf(text,"Health           %3d",Play->Health);
   mvaddstr(WantAntique ? 7 : 5,10,text);
   sprintf(text,"Bank %15s",FormatPrice(Play->Bank,num));
   mvaddstr(WantAntique ? 9 : 6,10,text);
   if (Play->Debt>0) attrset(DebtAttr);
   sprintf(text,"Debt %15s",FormatPrice(Play->Debt,num));
   mvaddstr(WantAntique ? 11 : 7,10,text);
   attrset(TitleAttr);
   if (WantAntique) sprintf(text,"Space %6d",Play->CoatSize);
   else {
      sprintf(text,"Bitches %3d  Space %6d",Play->Bitches,
              Play->CoatSize);
   }
   mvaddstr(0,78-strlen(text),text);
   print_location(Location[(int)Play->IsAt].Name);
   attrset(StatsAttr);

   c=0;
   if (DispDrug) {
      if (WantAntique) mvaddstr(1,54,"Trenchcoat");
      else mvaddstr(1,56,"Drugs");
      for (i=0;i<NUMDRUG;i++) {
         if (Play->Carry.Drug[i]>0) {
            sprintf(text,"%-7s  %3d",Drug[i].Name,Play->Carry.Drug[i]);
            mvaddstr(3+c/2,43+(c%2)*17,text);
            c++;
         }
      } 
   } else {
      mvaddstr(1,56,"Guns"); addch(ACS_HLINE);
      for (i=0;i<NUMGUN;i++) {
         if (Play->Carry.Gun[i]>0) {
            sprintf(text,"%-22s %3d",Gun[i].Name,Play->Carry.Gun[i]);
            mvaddstr(3+c,43,text);
            c++;
         }
      }
   }
   attrset(TextAttr);
}

void DisplaySpyReports(char *Data,struct PLAYER *From) {
/* Parses details about player "From" from string "Data" and then */
/* displays the lot, drugs and guns.                              */ 
   ReceivePlayerData(Data,From);

   clear_bottom();
   mvaddstr(17,1,"Spy reports for ");
   addstr(From->Name);

   mvaddstr(19,20,"Drugs...");
   print_status(From,1); nice_wait();
   clear_line(19);
   mvaddstr(19,20,"Guns...");
   print_status(From,0); nice_wait();

   print_status(Player,1); refresh();
}

int read_string(FILE *fp,char *buf,int maxlen) {
/* Reads a NULL-terminated string of maximum length "maxlen" into the */
/* buffer "buf" from file "fp". Returns 0 on success, EOF on failure. */
   int i,c;
   i=0;
   while (i<maxlen) {
      c=fgetc(fp);
      if (c==EOF || c==0) break;
      else buf[i++]=(char)c;
   }
   buf[i]=0;
   if (c==EOF) return EOF; else return 0;
}

void ClearInventory(Inventory *Inven) {
/* An inventory lists the numbers of guns and drugs carried by a player. */
/* This function simply clears the given inventory of all objects.       */
   int i;
   for (i=0;i<NUMGUN;i++) Inven->Gun[i]=0;
   for (i=0;i<NUMDRUG;i++) Inven->Drug[i]=0;
}

char IsInventoryClear(Inventory *Inven) {
/* Returns TRUE only if "Inven" contains no objects */
   int i;
   for (i=0;i<NUMGUN;i++) if (Inven->Gun[i] > 0) return FALSE;
   for (i=0;i<NUMDRUG;i++) if (Inven->Drug[i] > 0) return FALSE;
   return TRUE;
}

void AddInventory(Inventory *Cumul,Inventory *Add) {
/* Adds inventory "Add" into the contents of inventory "Cumul" */
   int i;
   for (i=0;i<NUMGUN;i++) Cumul->Gun[i]+=Add->Gun[i];
   for (i=0;i<NUMDRUG;i++) Cumul->Drug[i]+=Add->Drug[i];
}

void TruncateInventoryFor(Inventory *Inven,struct PLAYER *Play) {
/* Removes items from "Inven" if necessary such that player "Play" is */
/* able to carry them all. The cheapest objects are discarded.        */
   int i,Total,CheapIndex;
   int CheapestGun;
   Total=0;
   for (i=0;i<NUMGUN;i++) Total+=Inven->Gun[i];
   Total+=TotalGunsCarried(Play);
   while (Total > Play->Bitches+2) {
      CheapIndex=-1;  
      for (i=0;i<NUMGUN;i++) if (Inven->Gun[i] && (CheapIndex==-1 || 
                                 Gun[i].Price <= Gun[CheapIndex].Price)) {
         CheapIndex=i;
      }
      i=Total-Play->Bitches-2;
      if (Inven->Gun[CheapIndex] > i) {
         Inven->Gun[CheapIndex]-=i; Total-=i;
      } else {
         Total-=Inven->Gun[CheapIndex]; Inven->Gun[CheapIndex]=0;
      }
   }

   Total=Play->CoatSize;
   for (i=0;i<NUMGUN;i++) Total-=Inven->Gun[i]*Gun[i].Space;
   for (i=0;i<NUMDRUG;i++) Total-=Inven->Drug[i];
   while (Total < 0) {
      CheapestGun=-1;
      CheapIndex=-1;
      for (i=0;i<NUMGUN;i++) if (Inven->Gun[i] && (CheapIndex==-1 ||
                                 Gun[i].Price <= Gun[CheapIndex].Price)) {
         CheapIndex=i; CheapestGun=Gun[i].Price/Gun[i].Space;
      }
      for (i=0;i<NUMDRUG;i++) if (Inven->Drug[i] &&
         (CheapIndex==-1 ||
          (CheapestGun==-1 && Drug[i].MinPrice<=Drug[CheapIndex].MinPrice) ||
          (CheapestGun>=0 && Drug[i].MinPrice<=CheapestGun))) {
         CheapIndex=i; CheapestGun=-1;
      }
      if (CheapestGun>=0) { 
         Inven->Gun[CheapIndex]--;  
         Total+=Gun[CheapIndex].Space;
      } else {
         if (Inven->Drug[CheapIndex] >= -Total) {
            Inven->Drug[CheapIndex] += Total; Total=0;
         } else {
            Total+=Inven->Drug[CheapIndex]; Inven->Drug[CheapIndex]=0;
         }
      }
   }
} 

void PrintInventory(Inventory *Inven) {
/* Displays the contents of "Inven" at the bottom of the screen */
   int i,c;
   char text[80];
   c=0;
   attrset(TextAttr);
   for (i=0;i<NUMGUN;i++) if (Inven->Gun[i]) {
      sprintf(text,"%d %s",Inven->Gun[i],Gun[i].Name);
      mvaddstr(18+c/3,2+(c%3)*25,text);
      c++;
   }
   for (i=0;i<NUMDRUG;i++) if (Inven->Drug[i]) {
      sprintf(text,"%d %s",Inven->Drug[i],Drug[i].Name);
      mvaddstr(18+c/3,2+(c%3)*25,text);
      c++;
   }
}

int IsCarryingRandom(struct PLAYER *Play,int amount) {
/* Returns an index into the drugs array of a random drug that "Play" is */
/* carrying at least "amount" of. If no suitable drug is found after 5   */
/* attempts, returns -1.                                                 */ 
   int i,ind;
   for (i=0;i<5;i++) {
      ind=brandom(0,NUMDRUG);
      if (Play->Carry.Drug[ind] >= amount) {
         return ind;
      }
   }
   return -1;
}

int want_to_quit() {
/* Asks the user if he/she _really_ wants to quit dopewars */
   attrset(TextAttr);
   clear_line(22);
   attrset(PromptAttr);
   mvaddstr(22,1,"Are you sure you want to quit? ");
   attrset(TextAttr);
   return (GetKey("YN")!='N');
}

struct PLAYER *ListPlayers(struct PLAYER *Play,char Select,char *Prompt) {
/* Displays the "Prompt" if non-NULL, and then lists all clients     */
/* currently playing dopewars, other than the current player "Play". */
/* If "Select" is TRUE, gives each player a letter and asks the user */
/* to select one, which is returned by the function.                 */
   struct PLAYER *tmp;
   char text[80];
   int i,c;

   attrset(TextAttr);
   clear_bottom();
   if (!FirstClient || (!FirstClient->Next && 
       strcmp(FirstClient->Name,Play->Name)==0)) {
      mvaddstr(18,15,"No other players are currently logged on!");
      nice_wait();
      return 0;
   }
   mvaddstr(16,1,"Players currently logged on:-");

   i=0;
   for (tmp=FirstClient;tmp;tmp=tmp->Next) {
      if (strcmp(tmp->Name,Play->Name)==0) continue;
      if (Select) sprintf(text,"%c. %s",'A'+i,tmp->Name);
      else strcpy(text,tmp->Name);
      mvaddstr(17+i/3,(i%3)*36+3,text);
      i++;
   }

   if (Prompt) mvaddstr(22,10,Prompt);
   if (Select) {
      curs_set(1);
      attrset(TextAttr);
      c=0;
      while (c<'A' || c>='A'+i) { c=getch(); c=toupper(c); }
      if (Prompt) addch(c);
      tmp=FirstClient;
      while (c>='A') {
         if (tmp!=FirstClient) tmp=tmp->Next;
         while (strcmp(tmp->Name,Play->Name)==0) tmp=tmp->Next;
         c--;
      }
      return tmp;
   } else {
      nice_wait();
   }
   return NULL;
}

void change_name(char nullname) {
/* Prompts the user to change his or her name, and notifies the server */
   char NewName[80];
   nice_input("New name: ",NewName,30,23,0,0,0);
   if (NewName[0]) {
      SendClientMessage(nullname ? NULL : Player,C_NONE,"name",NULL,NewName);
      strcpy(Player->Name,NewName);
   }
}

void GiveErrand(struct PLAYER *Play) {
/* Prompts the user (player "Play") to give an errand to one of his/her */
/* bitches. The decision is relayed to the server for implementation.   */
   int c;
   char num[80];
   struct PLAYER *To;
   attrset(TextAttr);
   clear_bottom();
   mvaddstr(17,1,"Choose an errand to give one of your bitches...");
   attrset(PromptAttr);
   mvaddstr(18,5,"S>py on another drug dealer              (cost: ");
   addstr(FormatPrice(SPYPRICE,num)); addch(')');
   mvaddstr(19,5,"T>ip off the cops to another drug dealer (cost: ");
   addstr(FormatPrice(TIPPRICE,num)); addch(')');
   mvaddstr(20,5,"G>et stuffed");
   mvaddstr(21,2,"or N>o errand ? ");
   c = GetKey("STGN");
   switch (c) {
      case 'G':
         attrset(PromptAttr);
         addstr(" Are you sure? ");
         c=GetKey("YN");
         if (c=='Y') SendClientMessage(Play,C_NONE,"sackbitch",NULL,NULL);
         break;
      case 'T':
         To=ListPlayers(Play,TRUE,"Whom do you want to tip the cops off to? ");
         if (To) SendClientMessage(Play,C_NONE,"tipoff",To,NULL);
         break;
      case 'S':
         To=ListPlayers(Play,TRUE,"Whom do you want to spy on? ");
         if (To) SendClientMessage(Play,C_NONE,"spyon",To,NULL);
         break;
   }
}

void ClearMessageArea() {
/* Clears the message area */
   int i,j;
   if (WantAntique) return;
   for (i=10;i<=15;i++) {
      mvaddch(i,0,' ' | TextAttr);
      for (j=1;j<79;j++) addch(' ' | StatsAttr);
      addch(' ' | TextAttr);
   }
}

void do_game(struct PLAYER *Play) {
/* Loop which handles the user playing an interactive game (i.e. "Play" */
/* is a client connected to a server, either locally or remotely)       */
/* dopewars is essentially server-driven, so this loop simply has to    */
/* make the screen look pretty, respond to user keypresses, and react   */
/* to messages from the server.                                         */
   char text[BUFLEN],*pt,buf[BUFLEN];
   int i,c;
   char IsCarrying;
   fd_set readfs;
   int NumDrugsHere;
   struct PLAYER *tmp;
   InitialisePlayer(Play);
   attrset(TextAttr); clear_screen();
   ClearMessageArea();
   print_status(Play,1);

   attrset(TextAttr);
   clear_bottom();
   buf[0]=0;
   while (buf[0]==0) nice_input("Hey dude, what's your name? ",
                                buf,30,17,1,0,0);
   if (WantNetwork) {
      clear_line(17);
      mvaddstr(17,1,"Please wait... attempting to contact dopewars server...");
      refresh();
      pt=text;
      while (pt) {
         pt=SetupNetwork();
         if (pt) {
            attrset(TextAttr);
            clear_line(17);
            mvaddstr(17,1,"Could not start multiplayer dopewars (");
            addstr(pt);
            addstr(") ");
            attrset(PromptAttr);
            mvaddstr(18,1,"Will you... C>onnect to a different \
host and/or port");
            mvaddstr(19,1,"            Q>uit (where you can \
start a server by typing dopewars -s & )");
            mvaddstr(20,1,"         or P>lay single-player ? ");
            attrset(TextAttr);
            c=GetKey("CQP");
            if (c=='Q') { end_curses(); exit(1);}
            else if (c=='P') break;
            else {
               clear_bottom();
               nice_input("Hostname: ",ServerName,60,17,1,0,1);
               sprintf(text,"%d",Port);
               nice_input("Port: ",text,10,18,1,1,1);
               Port=atoi(text);
            }
         }
      }
   }
   SendClientMessage(NULL,C_NONE,"name",NULL,buf);
   strcpy(Play->Name,buf);

   while (1) {
      IsCarrying=0;
      for (i=0;i<NUMDRUG;i++) if (Play->Carry.Drug[i]>0) {
         IsCarrying=1; break;
      }

      attrset(TextAttr);
      c=0;
      for (i=0;i<NUMDRUG;i++) if (Play->Here->Price[i]>0) c++;
      if (c>0) {
         clear_bottom();
         mvaddstr(16,1,"Hey dude, the prices of drugs here are:");
      }
      c=0;
      for (i=0;i<NUMDRUG;i++) {
         if (Play->Here->Price[i] > 0) {
            sprintf(text,"%c. %-7s  %8s",'A'+c,Drug[i].Name,
                    FormatPrice(Play->Here->Price[i],buf));
            mvaddstr(17+c/3,(c%3)*23+8,text);
            c++;
         }
      }
      NumDrugsHere=c;
 
      attrset(PromptAttr);
      strcpy(text,"Will you B>uy");
      if (IsCarrying) strcat(text,", S>ell");
      if (Network) strcat(text,", T>alk, P>age, L>ist");
      if (!WantAntique && Play->Bitches>0) {
         strcat(text,", G>ive");
      }
      if (Play->Flags & SPYINGON) {
         strcat(text,", C>ontact");
      }
      if (Play->Flags & FIGHTING) {
         strcat(text,", R>un");
      } else {
         strcat(text,", J>et");
      }
      strcat(text,", or Q>uit? ");
      if (NumDrugsHere > 0) {
         mvaddstr(22,40-strlen(text)/2,text);
         attrset(TextAttr);
         curs_set(1);
      }
      refresh();
      while (1) {
         FD_ZERO(&readfs);
         FD_SET(0,&readfs);
         if (Client) FD_SET(ClientSock,&readfs);
         i=0;
         while (CheckWaitingClientMessage(Play,text)) {
            if (HandleClientMessage(text,Play)) return;
            i=1;
         }
         if (i) break;
         if (select(ClientSock+1,&readfs,NULL,NULL,NULL)==-1) {
            if (errno==EINTR) continue;
            perror("select"); exit(1);
         }
         if (Client && FD_ISSET(ClientSock,&readfs)) {
            pt=bgets(text,BUFLEN,ClientSock);
            chomp(text);
            clear_line(22);
            if (!pt) {
               mvaddstr(22,0,"Connection to server lost!");
               nice_wait();
               ShutdownNetwork();
               break;
            } else {
                if (HandleClientMessage(text,Play)) return;
                break;
            }
         } else if (FD_ISSET(0,&readfs)) {
            c=getch(); c=toupper(c);
            if ((c=='J' && !(Play->Flags&FIGHTING)) ||
                (c=='R' && Play->Flags&FIGHTING)) {
               jet(Play);
               break;
            } else if (c=='B') {
               DealDrugs(Play,TRUE);
               break;
            } else if (c=='S' && IsCarrying) {
               DealDrugs(Play,FALSE);
               break;
            } else if (c=='G' && !WantAntique && Play->Bitches>0) {
               GiveErrand(Play); break;
            } else if (c=='Q') {
               if (want_to_quit()==1) {
                  for (i=0;i<NUMDRUG;i++) Play->Here->Price[i]=0;
                  clear_bottom();
                  SendClientMessage(Play,C_NONE,"wantquit",NULL,NULL);
               }
               break;
            } else if (c=='L' && Network) {
               ListPlayers(Play,FALSE,NULL);
               break;
            } else if (c=='P' && Network) {
               tmp=ListPlayers(Play,TRUE,
                   "Whom do you want to page (talk privately to) ? ");
               if (tmp) {
                  nice_input("Talk: ",text,60,23,0,0,0);
                  if (text[0]) {
                     SendClientMessage(Play,C_NONE,"msgto",tmp,text);
                     sprintf(buf,"%s->%s: %s",Play->Name,tmp->Name,text);
                     display_message(buf);
                  }
               }
               break;
            } else if (c=='C' && Play->Flags & SPYINGON) {
               SendClientMessage(Play,C_NONE,"contactspy",NULL,NULL);
               break;
            } else if (c=='T' && Client) {
               nice_input("Talk: ",text,60,23,0,0,0);
               if (text[0]) {
                  SendClientMessage(Play,C_NONE,"msg",NULL,text);
                  sprintf(buf,"%s: %s",Play->Name,text);
                  display_message(buf);
               }
               break;
            } else {
               if (touchwin(stdscr)==ERR) printf("oops 1\n");
               if (refresh()==ERR) printf("oops 2\n");
            }
         }
      }
      curs_set(0);
      if (Play->Health==0) return;
   }
}

void InitList(DopeList *List) {
/* A DopeList is akin to a Vector class; it is a list of DopeEntry   */
/* structures, which can be dynamically extended or compressed. This */
/* function initialises the newly-created list pointed to by "List"  */
/* (A DopeEntry contains a PLAYER pointer and a counter, and is used */
/* by the server to keep track of tipoffs and spies.)                */
   List->Data=NULL;
   List->Number=0;
}

void ClearList(DopeList *List) {
/* Clears the list pointed to by "List" */
   free(List->Data);
   InitList(List);
}

void AddListEntry(DopeList *List,DopeEntry *NewEntry) {
/* Adds a new DopeEntry (pointed to by "NewEntry") to the list "List". */
/* A copy of NewEntry is placed into the list, so the original         */
/* structure pointed to by NewEntry can be reused.                     */
   if (!NewEntry || !List) return;
   List->Number++;
   List->Data = (DopeEntry *)realloc(List->Data,List->Number*
                                                sizeof(DopeEntry));
   if (!List->Data) {
      puts("ERROR: realloc failed!");
      exit(1);
   }
   memcpy(&(List->Data[List->Number-1]),NewEntry,sizeof(DopeEntry));
}

void RemoveListEntry(DopeList *List,int Index) {
/* Removes the DopeEntry at index "Index" from list "List" */
   if (!List || Index<0 || Index>=List->Number) return;

   memmove(&(List->Data[Index]),&(List->Data[Index+1]),
           (List->Number-1-Index)*sizeof(DopeEntry));
   List->Number--;
   List->Data = (DopeEntry *)realloc(List->Data,List->Number*
                                                sizeof(DopeEntry));
   if (List->Number==0) List->Data=NULL;
}

int GetListEntry(DopeList *List,struct PLAYER *Play) {
/* Returns the index of the DopeEntry matching "Play" in list "List" */
/* or -1 if this is not found.                                       */
   int i;
   for (i=List->Number-1;i>=0;i--) {
      if (List->Data[i].Play==Play) return i;
   }
   return -1;
}

void RemoveListPlayer(DopeList *List,struct PLAYER *Play) {
/* Removes (if it exists) the DopeEntry in list "List" matching "Play" */
   RemoveListEntry(List,GetListEntry(List,Play));
}

int ReadNameValue(FILE *fp,char *name,int namelen,
                  char *value,int valuelen) {
/* Reads a name=value pair from file "fp" into "name" and "value". Only   */
/* the first namelen and valuelen respectively bytes of the two buffers   */
/* are touched. Returns 0 on success, -1 on EOF, -2 if a comment is read, */
/* or -3 if the line does not contain a name=value pair.                  */
   char line[900],*pt;
   if (fgets(line,900,fp)==NULL) return -1;
   if (line[0]=='#') return -2;

   chomp(line); chomp(line);

   pt=index(line,'=');
   if (!pt) return -3;
   *pt='\0';
   strncpy(name,line,namelen);
   strncpy(value,pt+1,valuelen);
   name[namelen-1]=value[valuelen-1]='\0';
   return 0;
}

void SetupParameters(char *HiScoreFile,char *ServerName,int *Port) {
/* Initialises the location of the high score file, the name of the   */
/* host running a dopewars server, and the port on which the server   */
/* is to be found. Defaults can be overridden by the file ~/.dopewars */
/* or by specifying command line options.                             */
   FILE *fp;
   char name[80],value[80],ConfigFile[400],*pt;
   strcpy(HiScoreFile,"/var/lib/games/dopewars.sco");
   strcpy(ServerName,"localhost");
   *Port=7902;
   pt=getenv("HOME");
   if (!pt) return;
   if (strlen(pt) > 370) {
      printf("Error: home directory %s too long.\n",pt);
      return;
   }
   sprintf(ConfigFile,"%s/.dopewars",pt);
/*   printf("Attempting to open %s\n",ConfigFile); */
   fp=fopen(ConfigFile,"r");
   if (fp) {
      while (!feof(fp)) {
         if (ReadNameValue(fp,name,80,value,80)==0) {
            if (strcasecmp(name,"ScoreFile")==0) {
/*               printf("HiScore file is %s\n",value); */
               strcpy(HiScoreFile,value);
            } else if (strcasecmp(name,"Server")==0) {
/*               printf("Server is %s\n",value); */
               strcpy(ServerName,value);
            } else if (strcasecmp(name,"Port")==0) {
               *Port=atoi(value);
            } else {
               printf("Error: Unknown option %s in config file %s.\n",
                      name,ConfigFile);
            }
         }
      }
      fclose(fp);
   }
}

int main(int argc,char *argv[]) {
/* Main program entry point. Initialises variables and parses command line */
/* options, before starting a client, server, or AI player.                */
   struct timeval tp;
   int c;
   gettimeofday(&tp,NULL);
   srand(tp.tv_usec);
   FirstClient=FirstServer=NULL;
   SetupParameters(HiScoreFile,ServerName,&Port);
   strcpy(Noone.Name,"Noone");
   WantColour=WantNetwork=1; WantHelp=WantVersion=WantAntique=0;
   Server=AIPlayer=FALSE;
   while (1) {
      c=getopt(argc,argv,"anbchvf:o:sp:");
      if (c==EOF) break;
      switch(c) {
         case 'n': WantNetwork=0; break;
         case 'b': WantColour=0; break;
         case 'c': AIPlayer=1; break;
         case 'a': WantAntique=1; WantNetwork=0; break;
         case 'v': WantVersion=1; break;
         case 'h':
         case '?': WantHelp=1; break;
         case 'f': strcpy(HiScoreFile,optarg); break;
         case 'o': strcpy(ServerName,optarg); break;
         case 's': Server=TRUE; break;
         case 'p': Port=atoi(optarg); break;
      }
   }
   if (WantVersion) {
      printf("dopewars version %d.%d.%s\n",VERSION,MAJVER,MINVER);
      return 0;
   }
   if (WantHelp) {
      printf("dopewars version %d.%d.%s\n",VERSION,MAJVER,MINVER);
      puts("Usage: dopewars [OPTION]...");
      puts("Drug dealing game based on \"Drug Wars\" by John E. Dell\n");
      
      puts("  -b       \"black and white\" - i.e. do not use pretty colours");
      puts("              (by default colours are used where the \
terminal supports them)");
      puts("  -n       be boring and don't connect to any available \
dopewars servers");
      puts("              (i.e. single player mode)");
      puts("  -a       \"antique\" dopewars - keep as closely to the \
original version as");
      puts("              possible (this also disables any networking)");
      puts("  -f file  specify a file to use as the high score table");
      puts("              (by default /var/tmp/dopewars.sco is used)");
      puts("  -o addr  specify a hostname where the server for \
multiplayer dopewars ");
      puts("              can be found (in human-readable - e.g. \
nowhere.com - format)");
      puts("  -s       run in server mode (note: for a \"non-interactive\" \
server, simply");
      puts("              run as dopewars -s < /dev/null > /dev/null )");
      puts("  -p       specify the network port to use (default: 7902)");
      puts("  -c       create and run a computer player");
      puts("  -h       display this help information");
      puts("  -v       output version information and exit\n");

      puts("dopewars is Copyright (C) Ben Webb 1998-9, and released under the \
GNU GPL");
      puts("Report bugs to the author at ben@bellatrix.pcl.ox.ac.uk");
      return 0;
   }
   NUMLOCATION = (WantAntique ? 6 : MAXLOCATION);
   if (Server) {
      WantAntique=0;
      ServerLoop();
   } else if (AIPlayer) {
      AIPlayerLoop();
   } else {
      start_curses();
      display_intro();

      c='Y';
      while(c=='Y') {
         Player=AddPlayer(0,&FirstClient);
         do_game(Player);
         ClearWaitingClientMessages();
         ShutdownNetwork();
         CleanUpServer();
         attrset(TextAttr);
         mvaddstr(23,20,"Play again? ");
         c=GetKey("YN");
      }
      end_curses();
   }
   return 0;
}


